#include <u.h>
#include <libc.h>
#include <venti.h>

/* score of a zero length block */
uchar	vtzeroscore[VtScoreSize] = {
	0xda, 0x39, 0xa3, 0xee, 0x5e, 0x6b, 0x4b, 0x0d, 0x32, 0x55,
	0xbf, 0xef, 0x95, 0x60, 0x18, 0x90, 0xaf, 0xd8, 0x07, 0x09
};
