#include "stdinc.h"
#include "vac.h"
#include "dat.h"
#include "fns.h"
#include "error.h"

char ENoDir[] = "directory entry is not allocated";
char ENoFile[] = "no such file or directory";
char EBadPath[] = "bad path";
char EBadDir[] = "corrupted directory entry";
char EBadMeta[] = "corrupted meta data";
char ENotDir[] = "not a directory";
char ENotFile[] = "not a file";
char EIO[] = "i/o error";
char EBadOffset[] = "illegal offset";
char ETooBig[] = "file too big";
char EReadOnly[] = "read only";
char ERemoved[] = "file has been removed";
char ENilBlock[] = "illegal block address";
char ENotEmpty[] = "directory not empty";
char EExists[] = "file already exists";
char ERoot[] = "cannot remove root";
