/*
 * Somehow <errno.h> has been included on Mac OS X
 */
#undef EIO

extern char ENoDir[];
extern char EBadDir[];
extern char EBadMeta[];
extern char ENilBlock[];
extern char ENotDir[];
extern char ENotFile[];
extern char EIO[];
extern char EBadOffset[];
extern char ETooBig[];
extern char EReadOnly[];
extern char ERemoved[];
extern char ENotEmpty[];
extern char EExists[];
extern char ERoot[];
extern char ENoFile[];
extern char EBadPath[];
