extern char EBadAddr[];
extern char EBadDir[];
extern char EBadEntry[];
extern char EBadLabel[];
extern char EBadMeta[];
extern char EBadMode[];
extern char EBadOffset[];
extern char EBadPath[];
extern char EBadRoot[];
extern char EBadSuper[];
extern char EBlockTooBig[];
extern char ECacheFull[];
extern char EConvert[];
extern char EExists[];
extern char EFsFill[];
extern char EIO[];
extern char EInUse[];
extern char ELabelMismatch[];
extern char ENilBlock[];
extern char ENoDir[];
extern char ENoFile[];
extern char ENotDir[];
extern char ENotEmpty[];
extern char ENotFile[];
extern char EReadOnly[];
extern char ERemoved[];
extern char ENotArchived[];
extern char EResize[];
extern char ERoot[];
extern char ESnapOld[];
extern char ESnapRO[];
extern char ETooBig[];
extern char EVentiIO[];
